#include <cstdio>
#include <algorithm>

#include "groupscreen.h"
#include "fetch_group.h"
#include "fetch_max_score_for_song.h"
#include "fetch_max_score_for_player.h"
#include "fetch_needs_update.h"
#include "fetch_highscore.h"
#include "fonts.h"

GroupScreen::GroupScreen(pqxx::connection &conn, unsigned tournament, unsigned round, unsigned parallel)
	: tournament(tournament), round(round), parallel(parallel), scores_changed(conn, "scores"), conn(conn), valid(false)
{
}

GroupScreen::~GroupScreen()
{
}

bool GroupScreen::check_invalidated()
{
	if (!valid)
		return true;
	if (!scores_changed.get_flag())
		return false;

	bool needs_update;
	conn.perform(FetchNeedsUpdate(last_updated, tournament, round, parallel, &needs_update));

	if (!needs_update)
		scores_changed.reset_flag();
	
	return needs_update;
}

void GroupScreen::draw(unsigned char *buf)
{
	std::vector<TextDefer> td;
	
	scores_changed.reset_flag();

	Group group;
	conn.perform(FetchGroup(tournament, round, parallel, &group));
	gettimeofday(&last_updated, NULL);

	memset(buf, 0, 800 * 600 * 4);

	// main heading
	char heading[64];
	if (parallel == 0) {
		sprintf(heading, "Round %u", round);
	} else {
		sprintf(heading, "Round %u, Group %u", round, parallel);
	}

	{
		unsigned width = my_draw_text(heading, NULL, 48.0);
		my_draw_text_deferred(td, heading, 48.0, 800/2 - width/2, 60);
	}
	
	// Find out how wide each column has to be. First try unlimited width (ie.
	// long titles for everything); if that gets too long, try again with short
	// titles for chosen songs.
	unsigned width[16], num_scores;
	unsigned max_num_width = my_draw_text("8888", NULL, 22.0);
	unsigned mode, sumwidth;
	for (mode = 0; mode < 2; ++mode) {
		for (unsigned i = 0; i < 16; ++i)
			width[i] = 0;

		for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
			unsigned col = 1;
			width[0] = std::max(width[0], my_draw_text(i->nick, NULL, 18.0));

			for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
				if (j->chosen) {
					width[col] = std::max(width[col], my_draw_text((mode == 0) ? j->song.title : j->song.short_title, NULL, 12.0) + 
							max_num_width + 10);
				} else {		
					width[col] = std::max(width[col], my_draw_text(j->song.short_title, NULL, 12.0));
					width[col] = std::max(width[col], max_num_width);
				}
			}
		}

		num_scores = group.players[0].scores.size();

		width[num_scores + 1] = std::max(my_draw_text("Total", NULL, 12.0), max_num_width);
		width[num_scores + 2] = my_draw_text("Rank", NULL, 12.0);

		// if we're at long titles and that works, don't try the short ones
		sumwidth = 0;
			
		for (unsigned i = 0; i <= num_scores + 2; ++i)
			sumwidth += width[i] + 20;
			
		if (sumwidth < 780)
			break;
	}

	/* 
	 * If we have space to go, distribute as much as we can to the chosen song column, so we won't have
	 * total and rank jumping around.
	 */
	if (sumwidth < 780) {
		int first_chosen_col = -1;
		unsigned col = 1;

		for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i, ++col) {
			if (i->chosen) {
				first_chosen_col = col;
				break;
			}
		}

		if (first_chosen_col != -1) {
			printf("first_chosen_col=%u\n", first_chosen_col);
			width[first_chosen_col] += 780 - sumwidth;
		}
	}

	// make column headings from the first player's songs
	unsigned col = 1;
	unsigned x = 40 + width[0];
	for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i, ++col) {
		if (!i->chosen) {
			unsigned this_width = my_draw_text(i->song.short_title, NULL, 12.0);
			my_draw_text_deferred(td, i->song.short_title, 12.0, x + width[col] / 2 - this_width / 2, 100);
		}
		x += width[col] + 20;
	}

	my_draw_text_deferred(td, "Total", 12.0, x + width[num_scores + 1] / 2 - my_draw_text("Total", NULL, 12.0) / 2, 100);
	x += width[num_scores + 1] + 20;
	my_draw_text_deferred(td, "Rank", 12.0, x + width[num_scores + 2] / 2 - my_draw_text("Rank", NULL, 12.0) / 2, 100);
	
	// show all the players and the scores
	unsigned y = 140;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		my_draw_text_deferred(td, i->nick, 18.0, 20, y);

		unsigned x = 40 + width[0];

		unsigned col = 1;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			char text[16] = "";
			if (j->score != -1) {
				sprintf(text, "%u", j->score);
			}
	
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			if (j->chosen) {
				my_draw_text_deferred(td, text, 22.0, x + max_num_width - this_width, y);
				my_draw_text_deferred(td, (mode == 0) ? j->song.title : j->song.short_title, 12.0, x + max_num_width + 10, y);
			} else {
				my_draw_text_deferred(td, text, 22.0, x + width[col] / 2 - this_width / 2, y);
			}
			x += width[col] + 20;
		}

		// draw total
		{
			char text[16];
			sprintf(text, "%u", i->total);
			
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			my_draw_text_deferred(td, text, 22.0, x + width[num_scores + 1] / 2 - this_width / 2, y);
			x += width[num_scores + 1] + 20;
		}

		y += 40;
	}
	
	/*
	 * Approximate (but probably working quite well in practice) heuristic
	 * for finding the min and max rank of a player works as follows:
	 *
	 * First of all, find out, for each player in the group, what the
	 * maximum remaining score possibly can be (the minimum score is of
	 * course identical to the player's current total). For a random song,
	 * this is of course 1000 * (maximum feet rating) (but of course, that
	 * depends on whether we can play single or double! for now, assume
	 * double is okay, but this logic will be deferred to FetchMaxScore
	 * anyhow); for a random song, we simply pick the highest-ranking song
	 * we can find, EXCEPT those the player has chosen earlier AND the
	 * random songs this round, AND all random songs from elimination rounds
	 * (ie. rounds with only one group). (Phew!) This doesn't solve problems
	 * we'd face with more than one chosen song, but it should be good enough.
	 *
	 * After we've found the max and min scores for all players, it's a simple
	 * matter of sorting; the best attainable rank for player X is obtained if 
	 * X gets max score and all others get min score, the worst attainable rank
	 * is obtained if X gets min score and all others get max score.
	 *
	 * This is a bit SQL-heavy, but heck...
	 */
	std::vector<unsigned> max_score, min_score;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		unsigned min_score_tp = 0, max_score_tp = 0;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			if (j->score != -1) {
				// already given
				min_score_tp += j->score;
				max_score_tp += j->score;
			} else {
				unsigned max_score_this_song;
				if (j->song.id != -1) {
					// random song, or we know what song the player picked
					conn.perform(FetchMaxScoreForSong(tournament, j->song.id, &max_score_this_song));
				} else {
					conn.perform(FetchMaxScoreForPlayer(tournament, i->id, round, &max_score_this_song));
				}
				max_score_tp += max_score_this_song;
			}
		}
		max_score.push_back(max_score_tp);
		min_score.push_back(min_score_tp);
	}

	// now finally find min and max rank, and draw it all
	y = 140;
	for (unsigned i = 0; i < group.players.size(); ++i) {
		unsigned best_rank = 1, worst_rank = 1;
		for (unsigned j = 0; j < group.players.size(); ++j) {
			if (i == j)
				continue;

			if (max_score[i] < min_score[j])
				++best_rank;
			if (min_score[i] <= max_score[j])
				++worst_rank;
		}

		char text[16];
		if (best_rank == worst_rank)
			std::sprintf(text, "%u", best_rank);
		else
			std::sprintf(text, "%u-%u", best_rank, worst_rank);
		
		unsigned this_width = my_draw_text(text, NULL, 22.0);
		my_draw_text_deferred(td, text, 22.0, x + width[num_scores + 2] / 2 - this_width / 2, y);

		y += 40;
	}
		
	/*
	 * Next up (at the bottom) is "who's playing, what will he/she be playing, and
	 * optionally, how much to lead/win and how much to secure qualification" (the
	 * last one only in the final round). We assume playing is done in a modified
	 * zigzag; all the random songs are played first in zigzag/wrapping order (player
	 * 1 song 1, player 2 song 2, player 3 song 3, player 1 song 2, player 2 song 3,
	 * player 3 song 1, etc... assuming three songs and three players) and then all
	 * the chosen songs are played (we assume only one chosen song).
	 *
	 * The lines are as follows:
	 *
	 * <player>
	 * <song>
	 * High score: <hs> by <hsplayer> at <hsevent>
	 * Needs to lead: <leadscore>
	 * Needs to secure qualification: <qualscore>
	 * Needs to win group: <winscore>
	 */
	
	/* Find the first player with the fewest songs played. */
	unsigned min_played_songs = 9999, num_random_songs = 0;
	Player *next_player = NULL;
	for (std::vector<Player>::iterator i = group.players.begin(); i != group.players.end(); ++i) {
		unsigned this_played = 0, this_random_songs = 0;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			if (j->score != -1)
				++this_played;
			if (!j->chosen)
				++this_random_songs;
		}

		if (this_played < min_played_songs) {
			min_played_songs = this_played;
			next_player = &(*i);
			num_random_songs = this_random_songs;  // should be equal for all
		}
	}

	/* Find out what song this player is supposed to play next; try random songs first */ 
	Score *next_song = NULL;

	for (unsigned i = 0; i < num_random_songs; ++i) {
		unsigned j = (i + next_player->position - 1) % num_random_songs;
		if (next_player->scores[j].score == -1) {
			next_song = &(next_player->scores[j]);
			printf("Selecting score %u\n", j);
			break;
		}
	}

	// then all songs, if that didn't work out (slightly icky, but hey)
	if (next_song == NULL) {
		for (unsigned i = 0; i < num_scores; ++i) {
			unsigned j = (i + next_player->position) % num_scores;
			if (next_player->scores[j].score == -1) {
				next_song = &(next_player->scores[j]);
				printf("Selecting score %u\n", j);
				break;
			}
		}
	}

	if (next_song != NULL) {
		widestring text = widestring("Next player: ") + next_player->nick;
		unsigned this_width = my_draw_text(text, NULL, 24.0);
		my_draw_text(text, buf, 24.0, 400 - this_width/2, 420);

		if (next_song->song.id != -1) {
			this_width = my_draw_text(next_song->song.title, NULL, 20.0);
			my_draw_text(next_song->song.title, buf, 20.0, 400 - this_width/2, 457);

			Highscore hs;
			conn.perform(FetchHighscore(next_song->song.id, &hs));
			
			if (hs.score != -1) {
				text = widestring("High score: ") + widestring(pqxx::to_string(hs.score)) +
					widestring(", by ") + hs.nick + widestring(" in ") + hs.tournament_name;
				this_width = my_draw_text(text, NULL, 16.0);
				my_draw_text(text, buf, 16.0, 400 - this_width/2, 487);
			}
		}

		// only show lead/win/qualify for the last song
		if (min_played_songs == num_scores - 1) {
			/*
			 * Find out how much we need to lead, how much we need to be guaranteed
			 * to win the group, and how much we need to secure qualification. (FIXME:
			 * do the last one :-) )
			 */
			
			// find the best score we can get
			unsigned max_score_this_song;
			if (next_song->song.id != -1) {
				// random song, or we know what song the player picked
				conn.perform(FetchMaxScoreForSong(tournament, next_song->song.id, &max_score_this_song));
			} else {
				conn.perform(FetchMaxScoreForPlayer(tournament, next_player->id, round, &max_score_this_song));
			}

			unsigned y = 520;
			
			// see what score this player must beat to lead
			unsigned lead_beat = 0, win_beat = 0;
			for (unsigned i = 0; i < group.players.size(); ++i) {
				if (group.players[i].id == next_player->id)
					continue;
				
				lead_beat = std::max(lead_beat, group.players[i].total);
			}

			// find the best max score among the others
			for (unsigned i = 0; i < group.players.size(); ++i) {
				if (group.players[i].id == next_player->id)
					continue;

				win_beat = std::max(win_beat, max_score[i]);
			}

			/*
			 * There's a somewhat subtle point here. Normally, what a player would be interested in
			 * with regard to qualification would be a set of three values:
			 *
			 * 1. How much is the absolute minimum required to qualify, given that all others
			 *    fail?
			 * 2. How much will give a reasonable chance of qualifying, given the expected performance
			 *    of all the others?
			 * 3. How much will be enough to secure qualification, no matter what?
			 *
			 * Given perfect guessing, #2 would be "how much is needed to qualify"; however, it is
			 * completely impossible to give an exact value for that, and we're not into the guessing
			 * games. :-) #1 is often so low it's completely unrealistic (ie. far enough from #2 that
			 * it's not interesting), but #3, the most conservative estimate, is often a good measure.
			 * #3 is "how much is needed to _secure_ qualification", and that is usually what we
			 * print out when it's possible.
			 *
			 * However, in a few situations, #1 and #3 will be the exact same value, from which it
			 * follows (from the squeeze law, or just common sense :-) ) that #2 will be the same
			 * value as #1 and #3. (This usually happens near or at the end of a group.) In that
			 * case, we know the value we seek (ie. "how much is needed to qualify"), so we drop
			 * the word "secure" and just print it as-is.
			 *
			 * To find #1 and #3, we sort and pick out the values we need to beat in the best and
			 * the worst case.
			 */
			int qualify_beat_worst_case = -1, qualify_beat_best_case = -1;

			if (group.num_qualifying > 0) {
				std::vector<unsigned> tmp;
				
				for (unsigned i = 0; i < group.players.size(); ++i) {
					if (group.players[i].id == next_player->id)
						continue;
					tmp.push_back(max_score[i]);
				}
				std::sort(tmp.begin(), tmp.end());
				qualify_beat_worst_case = tmp[tmp.size() - group.num_qualifying];
				
				std::vector<unsigned> tmp2;
				for (unsigned i = 0; i < group.players.size(); ++i) {
					if (group.players[i].id == next_player->id)
						continue;
					tmp2.push_back(min_score[i]);
				}

				std::sort(tmp2.begin(), tmp2.end());
				qualify_beat_best_case = tmp2[tmp2.size() - group.num_qualifying];
			}
			
			// print out the lines we can attain
			if (next_player->total + max_score_this_song > lead_beat && (lead_beat != win_beat)) {
				int lead_need = std::max(lead_beat - next_player->total + 1, 0U);
				
				text = widestring("Needs to lead: ") + widestring(pqxx::to_string(lead_need));
				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, 400 - this_width/2, y);

				y += 30;
			}
			
			if (next_player->total + max_score_this_song > win_beat) {
				int win_need = std::max(win_beat - next_player->total + 1, 0U);
				
				text = widestring("Needs to win: ") + widestring(pqxx::to_string(win_need));

				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, 400 - this_width/2, y);

				y += 30;
			}

			if (next_player->total + max_score_this_song > qualify_beat_worst_case && (qualify_beat_worst_case != win_beat)) {
				int qual_need = std::max(qualify_beat_worst_case - next_player->total + 1, 0U);
				
				if (qualify_beat_worst_case == qualify_beat_best_case) {
					text = widestring("Needs to qualify: ") + widestring(pqxx::to_string(qual_need));
				} else {
					text = widestring("Needs to secure qualification: ") + widestring(pqxx::to_string(qual_need));
				}
				
				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, 400 - this_width/2, y);

				y += 30;
			}
		}
	}
	
	valid = true;
	draw_all_deferred_text(buf, td, last_text);
	last_text = td;
}

